class Model
  BONE_WEIGHTS = 64
  MAX_BONES_PER_MESH = 50

  vertexShaderSkinnedSource = [
    'attribute vec3 position;'
    'attribute vec2 texture;'
    'attribute vec3 normal;'
    'attribute vec3 weights;'
    'attribute vec3 bones;'

    'uniform mat4 viewMat;'
    'uniform mat4 modelMat;'
    'uniform mat4 projectionMat;'
    "uniform mat4 boneMat[#{MAX_BONES_PER_MESH}];"

    'uniform vec3 lightPos;'

    'varying vec2 vTexture;'
    'varying vec3 vNormal;'
    'varying vec3 vLightDir;'
    'varying vec3 vEyeDir;'

    'mat4 accumulateSkinMat() {'
      'mat4 result = weights.x * boneMat[int(bones.x)];'
      'result = result + weights.y * boneMat[int(bones.y)];'
      'result = result + weights.z * boneMat[int(bones.z)];'
      'return result;'
    '}'

    'mat3 getNormalMat(mat4 mat) {'
      'return mat3(mat[0][0], mat[1][0], mat[2][0], mat[0][1], mat[1][1], mat[2][1], mat[0][2], mat[1][2], mat[2][2]);'
    '}'

    'void main(void) {'
      'mat4 modelViewMat = viewMat * modelMat;'
      'mat4 skinMat = modelViewMat * accumulateSkinMat();'
      'mat3 normalMat = getNormalMat(skinMat);'

      'vec4 vPosition = skinMat * vec4(position, 1.0);'
      'gl_Position = projectionMat * vPosition;'

      'vTexture = texture;'
      'vNormal = normalize(normal * normalMat);'
      'vLightDir = normalize(lightPos - vPosition.xyz);'
      'vEyeDir = normalize(-vPosition.xyz);'
    '}'
  ].join('\n')

  vertexShaderStaticSource = [
    'attribute vec3 position;'
    'attribute vec2 texture;'
    'attribute vec3 normal;'

    'uniform mat4 viewMat;'
    'uniform mat4 modelMat;'
    'uniform mat4 projectionMat;'

    'uniform vec3 lightPos;'

    'varying vec2 vTexture;'
    'varying vec3 vNormal;'
    'varying vec3 vLightDir;'
    'varying vec3 vEyeDir;'

    'mat3 getNormalMat(mat4 mat) {'
      'return mat3(mat[0][0], mat[1][0], mat[2][0], mat[0][1], mat[1][1], mat[2][1], mat[0][2], mat[1][2], mat[2][2]);'
    '}'

    'void main(void) {',
      'mat4 modelViewMat = viewMat * modelMat;'
      'mat3 normalMat = getNormalMat(modelViewMat);'

      'vec4 vPosition = modelViewMat * vec4(position, 1.0);'
      'gl_Position = projectionMat * vPosition;'

      'vTexture = texture;'
      'vNormal = normalize(normal * normalMat);'
      'vLightDir = normalize(lightPos-vPosition.xyz);'
      'vEyeDir = normalize(-vPosition.xyz);'
    '}'
  ].join('\n')

  fragmentShaderSource = [
    'uniform sampler2D diffuse;'

    'varying vec2 vTexture;'
    'varying vec3 vNormal;'
    'varying vec3 vLightDir;'
    'varying vec3 vEyeDir;'

    'void main(void) {'
      'float shininess = 8.0;'
      'vec3 specularColor = vec3(0.1, 0.1, 0.1);'
      'vec3 lightColor = vec3(0.4, 0.4, 0.4);'
      'vec3 ambientLight = vec3(0.8, 0.8, 0.8);'

      'vec4 color = texture2D(diffuse, vTexture);'
      'vec3 normal = normalize(vNormal);'
      'vec3 lightDir = normalize(vLightDir);'
      'vec3 eyeDir = normalize(vEyeDir);'
      'vec3 reflectDir = reflect(-lightDir, normal);'

      'float specularLevel = color.a;'
      'float specularFactor = pow(clamp(dot(reflectDir, eyeDir), 0.0, 1.0), shininess) * specularLevel;'
      'float lightFactor = max(dot(lightDir, normal), 0.0);'
      'vec3 lightValue = ambientLight + (lightColor * lightFactor) + (specularColor * specularFactor);'
      'gl_FragColor = vec4(color.rgb * lightValue, 1.0);'
    '}'
  ].join('\n')

  constructor: ->
    @url = '/assets/models'

    @identityMat = mat4.create()
    mat4.identity(@identityMat)

    @complete = false
    @shader = null
    @meshes = null
    @vertexFormat = 0
    @vertexStride = 0
    @vertexBuffer = null
    @indexBuffer = null

    @bones = null
    @boneMatrices = null
    @dirtyBones = true

  load: (gl, name, @type) ->
    vertComplete = false
    modelComplete = false
    self = this

    # Load the binary model file
    request = new XMLHttpRequest()
    request.open('GET', "#{@url}/#{name}.wglvert", true)
    request.responseType = 'arraybuffer'

    request.onload = ->
      arrays = self.parseBinary(@response)
      self.compileBuffers(gl, arrays)
      vertComplete = true
      self.complete = true if modelComplete

    request.send(null)

    # Load the json model file
    request = new XMLHttpRequest()
    request.open('GET', "#{@url}/#{name}.wglmodel", true)

    request.onload = ->
      model = JSON.parse(@responseText)
      self.parseModel(model)
      self.compileMaterials(gl, self.meshes)
      modelComplete = true
      self.complete = true if vertComplete

    request.send(null)

    unless @shader
      attribs = ['position', 'texture', 'normal', 'weights', 'bones']
      uniforms = ['viewMat', 'modelMat', 'projectionMat', 'diffuse', 'lightPos', 'boneMat']
      if @type is 'skinned'
        @shader = @createShaderProgram(gl, vertexShaderSkinnedSource, fragmentShaderSource, attribs, uniforms)
      else
        @shader = @createShaderProgram(gl, vertexShaderStaticSource, fragmentShaderSource, attribs, uniforms)

  parseBinary: (buffer) ->
    output =
      vertexArray: null
      indexArray: null

    header = new Uint32Array(buffer, 0, 3)

    if @getLumpId(header[0]) isnt 'wglv'
      throw('Binary file magic number does not match expected value')

    if header[1] > 1
      throw('Binary file version is not supported')

    lumpCount = header[2]
    header = new Uint32Array(buffer, 12, lumpCount * 3)

    i = 0
    while i < lumpCount
      lumpId = @getLumpId(header[i * 3])
      offset = header[(i * 3) + 1]
      length = header[(i * 3) + 2]

      switch lumpId
        when 'vert'
          output.vertexArray = @parseVert(buffer, offset, length)
        when 'indx'
          output.indexArray = @parseIndex(buffer, offset, length)
      ++i

    if @vertexFormat & BONE_WEIGHTS
      @boneMatrices = new Float32Array(16 * MAX_BONES_PER_MESH)

    output

  parseVert: (buffer, offset, length) ->
    vertHeader = new Uint32Array(buffer, offset, 2)
    @vertexFormat = vertHeader[0]
    @vertexStride = vertHeader[1]

    new Uint8Array(buffer, offset + 8, length - 8)

  parseIndex: (buffer, offset, length) ->
    new Uint16Array(buffer, offset, length / 2)

  compileBuffers: (gl, arrays) ->
    @vertexBuffer = gl.createBuffer()
    gl.bindBuffer gl.ARRAY_BUFFER, @vertexBuffer
    gl.bufferData gl.ARRAY_BUFFER, arrays.vertexArray, gl.STATIC_DRAW

    @indexBuffer = gl.createBuffer()
    gl.bindBuffer gl.ELEMENT_ARRAY_BUFFER, @indexBuffer
    gl.bufferData gl.ELEMENT_ARRAY_BUFFER, arrays.indexArray, gl.STATIC_DRAW

  parseModel: (doc) ->
    @meshes = doc.meshes

    @bones = if doc.bones then doc.bones else []
    tempMat = mat4.create()

    # Force all bones to use efficient data structures
    for i, bone of @bones
      bone.pos = vec3.create(bone.pos)
      bone.rot = quat4.create(bone.rot)
      bone.bindPoseMat = mat4.create(bone.bindPoseMat)
      bone.boneMat = mat4.create()

      if bone.parent is -1
        bone.worldPos = bone.pos
        bone.worldRot = bone.rot
      else
        bone.worldPos = vec3.create()
        bone.worldRot = quat4.create()

  compileMaterials: (gl, meshes) ->
    for i, mesh of meshes
      mesh.diffuse = @loadTexture(gl, mesh.defaultTexture)

  getLumpId: (id) ->
    str = ''
    str += String.fromCharCode(id & 0xff)
    str += String.fromCharCode((id >> 8) & 0xff)
    str += String.fromCharCode((id >> 16) & 0xff)
    str += String.fromCharCode((id >> 24) & 0xff)
    str

  createShaderProgram: (gl, vertexShader, fragmentShader, attribs, uniforms) ->
    shaderProgram = gl.createProgram()

    vs = @compileShader(gl, vertexShader, gl.VERTEX_SHADER)
    fs = @compileShader(gl, fragmentShader, gl.FRAGMENT_SHADER)

    gl.attachShader(shaderProgram, vs)
    gl.attachShader(shaderProgram, fs)
    gl.linkProgram(shaderProgram)

    unless gl.getProgramParameter(shaderProgram, gl.LINK_STATUS)
      gl.deleteProgram(shaderProgram)
      gl.deleteShader(vs)
      gl.deleteShader(fs)
      return

    if attribs
      shaderProgram.attribute = {}
      for i, attrib of attribs
        shaderProgram.attribute[attrib] = gl.getAttribLocation(shaderProgram, attrib)

    if uniforms
      shaderProgram.uniform = {}
      for i, uniform of uniforms
        shaderProgram.uniform[uniform] = gl.getUniformLocation(shaderProgram, uniform)

    shaderProgram

  compileShader: (gl, source, type) ->
    shaderHeader = '#ifdef GL_ES\n'
    shaderHeader += 'precision highp float;\n'
    shaderHeader += '#endif\n'

    shader = gl.createShader(type)

    gl.shaderSource(shader, shaderHeader + source)
    gl.compileShader(shader)

    unless gl.getShaderParameter(shader, gl.COMPILE_STATUS)
      gl.deleteShader(shader)
      return

    shader

  loadTexture: (gl, src) ->
    texture = gl.createTexture()
    image = new Image()

    image.addEventListener 'load', ->
      gl.bindTexture(gl.TEXTURE_2D, texture)
      gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, image)
      gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR)
      gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR_MIPMAP_NEAREST)
      gl.generateMipmap(gl.TEXTURE_2D)

    image.src = src
    texture

  draw: (gl, viewMat, projectionMat) ->
    if @type is 'skinned'
      @drawSkinned(gl, viewMat, projectionMat)
    else
      @drawStatic(gl, viewMat, projectionMat)

  drawSkinned: (gl, viewMat, projectionMat) ->
    # Bind the appropriate buffers
    gl.bindBuffer(gl.ARRAY_BUFFER, @vertexBuffer)
    gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, @indexBuffer)

    gl.useProgram(@shader)

    gl.uniform3f(@shader.uniform.lightPos, 16, -32, 32)
    gl.uniformMatrix4fv(@shader.uniform.viewMat, false, viewMat)
    gl.uniformMatrix4fv(@shader.uniform.modelMat, false, @identityMat)
    gl.uniformMatrix4fv(@shader.uniform.projectionMat, false, projectionMat)

    gl.enableVertexAttribArray(@shader.attribute.position)
    gl.enableVertexAttribArray(@shader.attribute.texture)
    gl.enableVertexAttribArray(@shader.attribute.normal)
    gl.enableVertexAttribArray(@shader.attribute.weights)
    gl.enableVertexAttribArray(@shader.attribute.bones)

    # Setup the vertex layout
    gl.vertexAttribPointer(@shader.attribute.position, 3, gl.FLOAT, false, @vertexStride, 0)
    gl.vertexAttribPointer(@shader.attribute.texture, 2, gl.FLOAT, false, @vertexStride, 12)
    gl.vertexAttribPointer(@shader.attribute.normal, 3, gl.FLOAT, false, @vertexStride, 20)
    gl.vertexAttribPointer(@shader.attribute.weights, 3, gl.FLOAT, false, @vertexStride, 48)
    gl.vertexAttribPointer(@shader.attribute.bones, 3, gl.FLOAT, false, @vertexStride, 60)

    if @dirtyBones
      i = 0
      while i < @bones.length
        bone = @bones[i]
        @boneMatrices.set(bone.boneMat, i * 16)
        ++i

    for i, mesh of @meshes
      gl.activeTexture(gl.TEXTURE0)
      gl.bindTexture(gl.TEXTURE_2D, mesh.diffuse)
      gl.uniform1i(@shader.uniform.diffuse, 0)

      for j, submesh of mesh.submeshes
        if @bones.length > 0
          boneSet = @boneMatrices.subarray(submesh.boneOffset * 16, (submesh.boneOffset + submesh.boneCount) * 16)
          gl.uniformMatrix4fv(@shader.uniform.boneMat, false, boneSet)
        else
          gl.uniformMatrix4fv(@shader.uniform.modelMat, false, @identityMat)

        gl.drawElements(gl.TRIANGLES, submesh.indexCount, gl.UNSIGNED_SHORT, submesh.indexOffset * 2)

  drawStatic: (gl, viewMat, projectionMat) ->
    # Bind the appropriate buffers
    gl.bindBuffer(gl.ARRAY_BUFFER, @vertexBuffer)
    gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, @indexBuffer)

    gl.useProgram(@shader)

    gl.uniform3f(@shader.uniform.lightPos, 16, -32, 32)
    gl.uniformMatrix4fv(@shader.uniform.viewMat, false, viewMat)
    gl.uniformMatrix4fv(@shader.uniform.projectionMat, false, projectionMat)

    gl.enableVertexAttribArray(@shader.attribute.position)
    gl.enableVertexAttribArray(@shader.attribute.texture)
    gl.enableVertexAttribArray(@shader.attribute.normal)

    gl.vertexAttribPointer(@shader.attribute.position, 3, gl.FLOAT, false, @vertexStride, 0)
    gl.vertexAttribPointer(@shader.attribute.texture, 2, gl.FLOAT, false, @vertexStride, 12)
    gl.vertexAttribPointer(@shader.attribute.normal, 3, gl.FLOAT, false, @vertexStride, 20)

    for i, mesh of @meshes
      gl.activeTexture(gl.TEXTURE0)
      gl.bindTexture(gl.TEXTURE_2D, mesh.diffuse)
      gl.uniform1i(@shader.uniform.diffuse, 0)

      for j, submesh of mesh.submeshes
        gl.uniformMatrix4fv(@shader.uniform.modelMat, false, @identityMat)
        gl.drawElements(gl.TRIANGLES, submesh.indexCount, gl.UNSIGNED_SHORT, submesh.indexOffset * 2)

